{
  "type": "record",
  "name": "EthicalConsiderations",
  "namespace": "com.linkedin.pegasus2avro.ml.metadata",
  "doc": "This section is intended to demonstrate the ethical considerations that went into MLModel development, surfacing ethical challenges and solutions to stakeholders.",
  "fields": [
    {
      "name": "data",
      "type": [
        "null",
        {
          "type": "array",
          "items": "string"
        }
      ],
      "doc": "Does the MLModel use any sensitive data (e.g., protected classes)?",
      "default": null
    },
    {
      "name": "humanLife",
      "type": [
        "null",
        {
          "type": "array",
          "items": "string"
        }
      ],
      "doc": " Is the MLModel intended to inform decisions about matters central to human life or flourishing \u2013 e.g., health or safety? Or could it be used in such a way?",
      "default": null
    },
    {
      "name": "mitigations",
      "type": [
        "null",
        {
          "type": "array",
          "items": "string"
        }
      ],
      "doc": "What risk mitigation strategies were used during MLModel development?",
      "default": null
    },
    {
      "name": "risksAndHarms",
      "type": [
        "null",
        {
          "type": "array",
          "items": "string"
        }
      ],
      "doc": "What risks may be present in MLModel usage? Try to identify the potential recipients, likelihood, and magnitude of harms. If these cannot be determined, note that they were considered but remain unknown.",
      "default": null
    },
    {
      "name": "useCases",
      "type": [
        "null",
        {
          "type": "array",
          "items": "string"
        }
      ],
      "doc": "Are there any known MLModel use cases that are especially fraught? This may connect directly to the intended use section",
      "default": null
    }
  ],
  "Aspect": {
    "name": "mlModelEthicalConsiderations"
  }
}