{
  "type": "record",
  "name": "DataPlatformInstance",
  "namespace": "com.linkedin.pegasus2avro.common",
  "doc": "Tag aspect used for applying tags to an entity",
  "fields": [
    {
      "name": "platform",
      "type": "string",
      "doc": "Data Platform",
      "Searchable": {
        "addToFilters": true,
        "fieldType": "URN",
        "filterNameOverride": "Platform"
      },
      "java": {
        "class": "com.linkedin.pegasus2avro.common.urn.Urn"
      }
    },
    {
      "name": "instance",
      "type": [
        "null",
        "string"
      ],
      "doc": "Instance of the data platform (e.g. db instance)",
      "default": null,
      "Searchable": {
        "fieldName": "platformInstance",
        "fieldType": "URN"
      },
      "java": {
        "class": "com.linkedin.pegasus2avro.common.urn.Urn"
      }
    }
  ],
  "Aspect": {
    "name": "dataPlatformInstance"
  }
}